#include <stdio.h>
#include <string.h>

#include "mpi.h"

//! handles all the time measurement

static double timings[1025];
static int ids[1025];
static int epoch_used[5];
static int counter;

static char testname[50], method[50];
static char epoch_name[5][20];
static int bytes;
static MPI_File filehandle_values;

static int max_tests, current_tests;

//! sets some init parameter like the testname, method and number of bytes
//! also sets static the epoch names, the epoch_used array and the initial timing with a special id (zero)

void timing_init( char* ptestname, char* pmethod, int pbytes ) {

  int i;

  counter = 1;

  strncpy( testname, ptestname, 50 );
  strncpy( method, pmethod, 50 );
  bytes = pbytes;

  snprintf( &epoch_name[0][0], 20, "ddt_create_overhead");
  snprintf( &epoch_name[1][0], 20, "pack");
  snprintf( &epoch_name[2][0], 20, "communication");
  snprintf( &epoch_name[3][0], 20, "unpack");
  snprintf( &epoch_name[4][0], 20, "ddt_free_overhead");

  for( i = 0 ; i < 5 ; i++) {
    epoch_used[i] = 0;
  }

  ids[0] = 0;
  timings[0] = MPI_Wtime();
}

//! writes the timing values to the output file
//! if the last parameter is set, then all the unused epochs are also printed

void timing_print( int last ) {

  char line[256];
  int i;

  for( i=1; i<counter ; i++ ) {
//! if a epoch id is not in 1..5 is found, only the id is printed (instead of
//! the the epoch name)
    if ( (ids[i] > 5) || (ids[i] < 1) ) {
      snprintf(line, 256, "%30s%30s%15i%20i%20.4f\n", testname, method, bytes, ids[i], (timings[i]-timings[i-1])*1000000);
    } else {
      snprintf(line, 256, "%30s%30s%15i%20s%20.4f\n", testname, method, bytes, epoch_name[ids[i]-1], (timings[i]-timings[i-1])*1000000);
//! keep tracking if a epoch is used or not
      if (epoch_used[ids[i]-1] == 0) {
        epoch_used[ids[i]-1] = 1;
      }
    }
    MPI_File_write(filehandle_values, line, strlen(line), MPI_CHAR, MPI_STATUS_IGNORE);
  }

  if (last == 1) {
//! each epoch must appear for all tests to make the analysis afterwards easier
//! at the end of the test, we print for each unused epoch a line with a timing of zero micro seconds
    for( i=0; i<5 ; i++ ) {
      if (epoch_used[i] == 0) {
        snprintf(line, 256, "%30s%30s%15i%20s%20.4f\n", testname, method, bytes, epoch_name[i], (double) 0);
        MPI_File_write(filehandle_values, line, strlen(line), MPI_CHAR, MPI_STATUS_IGNORE);
        current_tests++;
      }
    }
  }

//! prints some progression information to stdout
  current_tests = current_tests + counter - 1;
  printf("Finished %7.3f%% of all tests\n", (double) current_tests/max_tests*100);
}

//! records the timing for the given epoch id
//! if number of recorded timings exceed 1024, then the function empties
//! the buffer and they will be printed

void timing_record( int id ) {

  timings[counter] = MPI_Wtime();
  ids[counter++] = id;
  if ( counter > 1024 ) {
    timing_print(0);
    ids[0] = 0;
    timings[0] = MPI_Wtime();
    counter = 1;
  }
}

//! opens a file handle, to where the timing values are written
void timing_open_file( char* filename ) {
      
  char line[256];
  int ier;

  ier = MPI_File_open( MPI_COMM_SELF, filename, MPI_MODE_EXCL + MPI_MODE_CREATE + MPI_MODE_WRONLY, MPI_INFO_NULL, &filehandle_values );

  if ( ier != MPI_SUCCESS ) {
    printf("Error at open file %s for writing the timing values. The file probably already exists.\nWill now abort.\n", filename);
    MPI_Abort( MPI_COMM_WORLD, 1 );
  }

  snprintf(line, 256, "%30s%30s%15s%20s%20s\n", "testname", "method", "bytes", "id", "time");
  MPI_File_write(filehandle_values, line, strlen(line), MPI_CHAR, MPI_STATUS_IGNORE);
}

//! closes the above mentioned filehandle
void timing_close_file() {
  MPI_File_close( &filehandle_values );
}

//! sets the initial parameter for the progression bar
void timing_set_max_tests( int value ) {

  max_tests = value;
  current_tests = 0;

}
